/*
 * jwm-menu-loader.c
 * 
 *  - Copyright (C) 2022 Aitor Cuadrado Zubizarreta <aitor_czr@gnuinos.org> 
 * 
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See the LISENCE file.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <dirent.h>
#include <signal.h>
#include <sys/inotify.h>
#include <assert.h>
#include <pstat/libpstat.h>
#include <sys/socket.h>
#include <sys/wait.h>  // waitpid()

#define MAX_SIZE 1024
#define ARRAY_DIM 1

const char *progname = "jwm-menu-loader";

// Function prototypes
void kill_all_instances();

static
void sig_handler(int signum, siginfo_t *info, void *extra)
{ 
   switch(signum) {
         
      case SIGTERM:
      case SIGINT:
      case SIGQUIT:
         exit(EXIT_SUCCESS);  /* exit gracefully. */
         break;
         
      default:
         break;
   }
}

/*====== A subprogram like popen() but with diferences explained below ======*/
/* The pipe is to read in the current process and connected to stderr in the */
/* new one. The intent is to report error messages. Plus pid is returned.    */
/* If dirpath is an absolute path, chdir to it prior to excuting command.    */
FILE *epopen(const char *command,  pid_t *pid)
{
  int pipefd[2];
  pid_t p_id;
  FILE *pf;
  int rc;

  /*--------------------------- Create the pipe -----------------------------*/
  if(pipe(pipefd))
    {
      *pid = -1;
      return NULL;
    }

  /*------------------------------------ fork -------------------------------*/
  p_id = fork();
  switch(p_id)
    {
    case 0: /* the child */
      close(pipefd[0]);
      close(STDERR_FILENO);
      dup2(pipefd[1], STDERR_FILENO);
      close(pipefd[1]);
      execl( "/bin/sh", "sh", "-c", command, (char *)NULL );
      fprintf(stderr, "%s: error invoking /bin/sh: %s\n", progname, strerror(errno));
      exit(EXIT_FAILURE); /* reminder: we are the child */
    case -1: /* fork() failed ! */
      rc = errno;
      close(pipefd[0]);
      close(pipefd[1]);
      errno = rc;
      *pid = -1;
      return NULL;
    default: /* the parent */
      close(pipefd[1]);
      pf = fdopen(pipefd[0], "r");
      *pid = p_id;
      return pf; /* The caller must close the pipe after use */
    }
}

// Usage: strncpy_t(str1, sizeof(str1), str2, strlen(str2));
// Copy string "in" with at most "insz" chars to buffer "out", which
// is "outsz" bytes long. The output is always 0-terminated. Unlike
// strncpy(), strncpy_t() does not zero fill remaining space in the
// output buffer:
char* strncpy_t(char* out, size_t outsz, const char* in, size_t insz)
{
   assert(outsz > 0);
   while(--outsz > 0 && insz > 0 && *in) { *out++ = *in++; insz--; }
   *out = 0;
   return out;
}

// thread-safe and async-safe int to string for base 10
static
void itoa10_safe( int val, char* str ) {
   
   int rc = 0;
   int i = 0;
   int len = 0;
   int j = 0;
   bool neg = false;
   
   // sign check
   if( val < 0 ) {
      val = -val;
      neg = true;
   }
   
   // consume, lowest-order to highest-order
   if( val == 0 ) {
      str[i] = '0';
      i++;
   }
   else {
      while( val > 0 ) {
      
         int r = val % 10;
      
         str[i] = '0' + r;
         i++;
      
         val /= 10;
      }
   }
   
   if( neg ) {
      
      str[i] = '-';
      i++;
   }
   
   len = i;
   i--;
   
   // reverse order to get the number 
   while( j < i ) {
      
      char tmp = *(str + i);
      *(str + i) = *(str + j);
      *(str + j) = tmp;
      
      j++;
      i--;
   }
   
   str[len] = '\0';
}

int main(int argc, char **argv)
{
    pid_t p_id, sid = 0;
	int rc = 0;
	int f = -1;
	int sockets[2];
	char msg[64];
   
    kill_all_instances();
	
	// At this point, any previous process has been killed;
	// if argc>1, then exit doing nothing more, whatever the values of the arguments in the command line
	// We aren't parsing any argument in cache-view so far
	if(argc>1)
		return 1;
   
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, sockets) < 0) {
		perror("opening stream socket pair");
		exit(1);
	}
	
    p_id = fork();
   
	if( p_id == 0 ) {   /* child */
	
		struct sigaction sa;
		
		int rc = 0;
		int inotfd = -1;
		int watch_usr = -1;
		int watch_usrlocal = -1; 
		int watch_xdg = -1;
		fd_set rfds;
      
        close(sockets[1]);
	
		memset (&sa, 0, sizeof(sa));
     
		/*---------------- initialize signal handling ------------ */
		sa.sa_sigaction = &sig_handler;
		sa.sa_flags = SA_RESETHAND | SA_NODEFER | SA_SIGINFO | SA_RESTART;
		sigemptyset(&sa.sa_mask);
		sigaction(SIGTERM, &sa, 0);
		sigaction(SIGINT,  &sa, 0);
		sigaction(SIGQUIT, &sa, 0);

		inotfd = inotify_init1(IN_CLOEXEC);
		if (inotfd < 0) {
			fprintf(stderr, "%s: cannot initialize inotify: %s\n", progname, strerror(errno) );
			exit(EXIT_FAILURE);
		}
    
		watch_usr = inotify_add_watch(inotfd, "/usr/share/applications", IN_CREATE | IN_DELETE | IN_MODIFY);
		if(watch_usr == -1) {

			fprintf(stderr, "%s: cannot add inotify watch '/usr/share/applications': %s\n", progname, strerror(errno));
			exit(EXIT_FAILURE);
		}
    
		watch_usrlocal = inotify_add_watch(inotfd, "/usr/local/share/applications", IN_CREATE | IN_DELETE | IN_MODIFY);
		if(watch_usrlocal == -1) {
	      
			// do not exit here
			//fprintf(stderr, "%s: cannot add inotify watch '/usr/local/share/applications': %s\n", progname, strerror(errno));
		}
      /*
		watch_xdg = inotify_add_watch(inotfd, "/etc/xdg/menus", IN_CREATE | IN_DELETE | IN_MODIFY);
		if(watch_xdg == -1) {

			fprintf(stderr, "cannot add inotify watch '/etc/xdg/menus': %s\n", progname, strerror(errno));
			exit(EXIT_FAILURE);
		}
      */  
      
        close(sockets[0]);
   
Select:			
		FD_ZERO(&rfds);
		FD_SET(inotfd, &rfds);
		
		rc = select(inotfd + 1, &rfds, NULL, NULL, NULL);
		if (rc < 0) {

			if (errno == EINTR) {
				rc = -errno;
				fprintf(stderr, "%s: select(): '%s' errno = %d\n", progname, strerror(errno), rc);
				goto Select;
			}
		}
   
		// reset the inotify watch descriptors 
		if( watch_usr != -1 ) {
				
			rc = inotify_rm_watch( inotfd, watch_usr );
			if( rc < 0 ) {
				rc = -errno;
				fprintf(stderr, "%s: close(events_wd %d) rc = %d", progname, watch_usr, rc );
			}
			else {
              
				watch_usr = -1;
			}
		}
   
		if( watch_usrlocal != -1 ) {
	    
			rc = inotify_rm_watch( inotfd, watch_usrlocal );
			if( rc < 0 ) {
         
				rc = -errno;
				fprintf(stderr, "%s: close(events_wd %d) rc = %d", progname, watch_usrlocal, rc );
			}
			else {
         
				watch_usrlocal = -1;
			}
		}
		
		if( watch_xdg != -1 ) {
				
			rc = inotify_rm_watch( inotfd, watch_xdg );
			if( rc < 0 ) {
				rc = -errno;
				fprintf(stderr, "%s: close(events_wd %d) rc = %d", progname, watch_xdg, rc );
			}
			else {
              
				watch_xdg = -1;
			}
		}		

		/* FD_ISSET(0, &rfds) will be true. */
		if ( FD_ISSET(inotfd, &rfds) ) {
	        
	        pid_t pid;
	        int wstatus;
            FILE *pfin = NULL;
                
            pfin = epopen("jwm -reload", &pid);
            if(pfin) { 
               fclose(pfin);
               waitpid(pid, &wstatus, 0);
   	        }
	
	        if(wstatus == 0) { 
	          
	          char *const params[] = {"jwm-menu-loader", (char*)0};
		      execv("/usr/bin/jwm-menu-loader", params);
		    }
		    else {
			  
			  exit(EXIT_FAILURE);
			}    
		}	  
	}
	
	else if(p_id == -1) {  /* fork() failed ! */
      
       rc = errno;
       
       
       errno = rc;
       exit(EXIT_FAILURE);
    }
    else { /* the parent */
      
       close(sockets[0]);
       close(sockets[1]);
       
       return 0;
    }
}

void kill_all_instances()
{
	DIR *dirp = NULL;
	struct dirent *dir = NULL;
    const char *processes[ARRAY_DIM] = { "jwm-menu-loader" };
    pid_t pid;
    char pidbuf[10];
    
    pid = getpid();
    pidbuf[9] = 0;
    itoa10_safe(pid, pidbuf);
	
	dirp=opendir("/proc");
	if(!dirp) {
		fprintf(stderr, "\n\tERROR: unable to open /proc for read\n\n", progname);
		exit(EXIT_FAILURE);
	}
	
	while((dir=readdir(dirp)) != NULL ) {

        int rc = 0;
	    struct stat sb;
        char *proc_cmd = NULL;
        bool is_uint_ok = true;
        char pth[MAX_SIZE] = {0};
	    
	    for(unsigned i=0; i<strlen(dir->d_name); i++) {
		   if(!isdigit(dir->d_name[i])) is_uint_ok = false;
	    }
        
        if(!is_uint_ok) continue;
    
        strncpy_t(pth, sizeof(pth), "/proc/", strlen("/proc/"));
        strcat(pth, dir->d_name);
        if (stat(pth, &sb) == -1) {
           perror("stat");
           // do not exit, go ahead
           continue;
        }
        
        if(!strcmp(dir->d_name, pidbuf)) continue;
          
        if(!lstat(dir->d_name, &sb)) continue;
 
        /* Only consider directories */
        if(!S_ISDIR(sb.st_mode)) continue;
        
		rc = pstat_get_binary_path_from_string(&proc_cmd, MAX_SIZE, dir->d_name);
 
        for(unsigned int i=0; i<ARRAY_DIM; i++) {		                    
           if(proc_cmd && !strcmp(rindex(proc_cmd, '/') + 1, processes[i])) {
              if( (kill(atoi(dir->d_name), SIGTERM)) != 0 ) kill(atoi(dir->d_name), SIGKILL);
           }
        }

	    free(proc_cmd);
	}
	
    closedir(dirp);
}
